@@ -15,6 +15,7 @@
  */
 
 #include "system.h"
+#include "popcount.h"
 
 /* locally defined functions */
 static void segment_tell PROTO((uptr seg));
@@ -188,18 +189,21 @@ void S_immobilize_object(x) ptr x; {
     si = MaybeSegInfo(ptr_get_segment(x));
  
   if ((si != NULL) && (si->generation != static_generation)) {
-    tc_mutex_acquire()
+    tc_mutex_acquire();
 
     /* Try a little to to support cancellation of segment-level
      * immobilzation --- but we don't try too hard */
-    if (si->must_mark < MUST_MARK_INFINITY)
+    if (si->must_mark < MUST_MARK_INFINITY) {
       si->must_mark++;
+      if (si->generation == 0)
+        S_G.must_mark_gen0 = 1;
+    }
 
     /* Note: for `space_new`, `must_mark` doesn't really mean all
        objects must be marked; only those in the locked list must be
        marked. Non-locked objects on `space_new` cannot be immobilized. */
 
-    tc_mutex_release()
+    tc_mutex_release();
   }
 }
 
@@ -212,7 +216,7 @@ void S_mobilize_object(x) ptr x; {
     si = MaybeSegInfo(ptr_get_segment(x));
 
   if ((si != NULL) && (si->generation != static_generation)) {
-    tc_mutex_acquire()
+    tc_mutex_acquire();
 
     if (si->must_mark == 0)
       S_error_abort("S_mobilize_object(): object was definitely not immobilzed");
@@ -221,7 +225,7 @@ void S_mobilize_object(x) ptr x; {
     if (si->must_mark < MUST_MARK_INFINITY)
       --si->must_mark;
   
-    tc_mutex_release()
+    tc_mutex_release();
   }
 }
 
@@ -257,7 +261,7 @@ IBOOL Slocked_objectp(x) ptr x; {
 
   if (IMMEDIATE(x) || (si = MaybeSegInfo(ptr_get_segment(x))) == NULL || (g = si->generation) == static_generation) return 1;
 
-  tc_mutex_acquire()
+  tc_mutex_acquire();
 
   ans = 0;
   for (ls = S_G.locked_objects[g]; ls != Snil; ls = Scdr(ls)) {
@@ -267,7 +271,7 @@ IBOOL Slocked_objectp(x) ptr x; {
     }
   }
 
-  tc_mutex_release()
+  tc_mutex_release();
 
   return ans;
 }
@@ -275,7 +279,7 @@ IBOOL Slocked_objectp(x) ptr x; {
 ptr S_locked_objects(void) {
   IGEN g; ptr ans; ptr ls;
 
-  tc_mutex_acquire()
+  tc_mutex_acquire();
 
   ans = Snil;
   for (g = 0; g <= static_generation; INCRGEN(g)) {
@@ -284,7 +288,7 @@ ptr S_locked_objects(void) {
     }
   }
 
-  tc_mutex_release()
+  tc_mutex_release();
 
   return ans;
 }
@@ -294,19 +298,23 @@ void Slock_object(x) ptr x; {
 
  /* weed out pointers that won't be relocated */
   if (!IMMEDIATE(x) && (si = MaybeSegInfo(ptr_get_segment(x))) != NULL && (g = si->generation) != static_generation) {
-    tc_mutex_acquire()
+    ptr tc = get_thread_context();
+    tc_mutex_acquire();
     S_pants_down += 1;
     /* immobilize */
-    if (si->must_mark < MUST_MARK_INFINITY)
+    if (si->must_mark < MUST_MARK_INFINITY) {
       si->must_mark++;
+      if (si->generation == 0)
+        S_G.must_mark_gen0 = 1;
+    }
    /* add x to locked list. remove from unlocked list */
-    S_G.locked_objects[g] = S_cons_in((g == 0 ? space_new : space_impure), g, x, S_G.locked_objects[g]);
+    S_G.locked_objects[g] = S_cons_in(tc, (g == 0 ? space_new : space_impure), g, x, S_G.locked_objects[g]);
     if (S_G.enable_object_counts) {
       if (g != 0) S_G.countof[g][countof_pair] += 1;
     }
     (void)remove_first_nomorep(x, &S_G.unlocked_objects[g], 0);
     S_pants_down -= 1;
-    tc_mutex_release()
+    tc_mutex_release();
   }
 }
 
@@ -314,7 +322,8 @@ void Sunlock_object(x) ptr x; {
   seginfo *si; IGEN g;
 
   if (!IMMEDIATE(x) && (si = MaybeSegInfo(ptr_get_segment(x))) != NULL && (g = si->generation) != static_generation) {
-    tc_mutex_acquire()
+    ptr tc = get_thread_context();
+    tc_mutex_acquire();
     S_pants_down += 1;
     /* mobilize, if we haven't lost track */
     if (si->must_mark < MUST_MARK_INFINITY)
@@ -322,13 +331,13 @@ void Sunlock_object(x) ptr x; {
    /* remove first occurrence of x from locked list. if there are no
       others, add x to unlocked list */
     if (remove_first_nomorep(x, &S_G.locked_objects[g], (si->space == space_new) && (si->generation > 0))) {
-      S_G.unlocked_objects[g] = S_cons_in((g == 0 ? space_new : space_impure), g, x, S_G.unlocked_objects[g]);
+      S_G.unlocked_objects[g] = S_cons_in(tc, (g == 0 ? space_new : space_impure), g, x, S_G.unlocked_objects[g]);
       if (S_G.enable_object_counts) {
         if (g != 0) S_G.countof[g][countof_pair] += 1;
       }
     }
     S_pants_down -= 1;
-    tc_mutex_release()
+    tc_mutex_release();
   }
 }
 
@@ -347,7 +356,7 @@ ptr s_help_unregister_guardian(ptr *pls, ptr tconc, ptr result) {
 
 ptr S_unregister_guardian(ptr tconc) {
   ptr result, tc; IGEN g;
-  tc_mutex_acquire()
+  tc_mutex_acquire();
   tc = get_thread_context();
   /* in the interest of thread safety, gather entries only in the current thread, ignoring any others */
   result = s_help_unregister_guardian(&GUARDIANENTRIES(tc), tconc, Snil);
@@ -355,15 +364,15 @@ ptr S_unregister_guardian(ptr tconc) {
   for (g = 0; g <= static_generation; INCRGEN(g)) {
     result = s_help_unregister_guardian(&S_G.guardians[g], tconc, result);
   }
-  tc_mutex_release()
+  tc_mutex_release();
   return result;
 }
 
 #ifndef WIN32
 void S_register_child_process(INT child) {
-  tc_mutex_acquire()
+  tc_mutex_acquire();
   S_child_processes[0] = Scons(FIX(child), S_child_processes[0]);
-  tc_mutex_release()
+  tc_mutex_release();
 }
 #endif /* WIN32 */
 
@@ -378,7 +387,7 @@ void S_set_enable_object_counts(IBOOL eoc) {
 ptr S_object_counts(void) {
   IGEN grtd, g; ptr ls; iptr i; ptr outer_alist;
 
-  tc_mutex_acquire()
+  tc_mutex_acquire();
 
   outer_alist = Snil;
 
@@ -435,7 +444,7 @@ ptr S_object_counts(void) {
     }
   }
 
-  tc_mutex_release()
+  tc_mutex_release();
 
   return outer_alist;
 }
@@ -451,12 +460,12 @@ void S_set_enable_object_backreferences(IBOOL eoc) {
 ptr S_object_backreferences(void) {
   IGEN g; ptr ls = Snil;
 
-  tc_mutex_acquire()
+  tc_mutex_acquire();
 
   for (g = S_G.max_nonstatic_generation+1; g--; )
     ls = Scons(S_G.gcbackreference[g], ls);
 
-  tc_mutex_release()
+  tc_mutex_release();
 
   return ls;
 }
@@ -473,7 +482,7 @@ void Scompact_heap() {
   IBOOL eoc = S_G.enable_object_counts;
   S_pants_down += 1;
   S_G.enable_object_counts = 1;
-  S_gc_oce(tc, S_G.max_nonstatic_generation, static_generation, Sfalse);
+  S_gc_oce(tc, S_G.max_nonstatic_generation, static_generation, static_generation, Sfalse);
   S_G.enable_object_counts = eoc;
   S_pants_down -= 1;
 }
@@ -494,15 +503,25 @@ void Scompact_heap() {
    If noisy is nonzero, additional comments may be included in the output
 */
 
+#ifdef __MINGW32__
+# include <inttypes.h>
+# define PHtx "%" PRIxPTR
+# define Ptd "%" PRIdPTR
+#else
+# define PHtx "%#tx"
+# define Ptd "%td"
+#endif
+
 static void segment_tell(seg) uptr seg; {
   seginfo *si;
   ISPC s, s1;
   static char *spacename[max_space+1] = { alloc_space_names };
 
-  printf("segment %#tx", (ptrdiff_t)seg);
+  printf("segment "PHtx"", (ptrdiff_t)seg);
   if ((si = MaybeSegInfo(seg)) == NULL) {
     printf(" out of heap bounds\n");
   } else {
+    printf(" si=%p", si);
     printf(" generation=%d", si->generation);
     s = si->space;
     s1 = si->space;
@@ -527,13 +546,100 @@ void S_addr_tell(ptr p) {
   segment_tell(addr_get_segment(p));
 }
 
+static void check_pointer(ptr *pp, IBOOL address_is_meaningful, ptr base, uptr seg, ISPC s, IBOOL aftergc) {
+  ptr p = *pp;
+  if (!IMMEDIATE(p)) {
+    seginfo *psi = MaybeSegInfo(ptr_get_segment(p));
+    if (psi != NULL) {
+      if ((psi->space == space_empty)
+          || psi->old_space
+          || (psi->marked_mask && !(psi->marked_mask[segment_bitmap_byte(p)] & segment_bitmap_bit(p))
+              /* corner case: a continuation in space_count_pure can refer to code via CLOSENTRY
+                 where the entry point doesn't have a mark bit: */
+              && !((s == space_count_pure) && (psi->space == space_code)))) {
+        S_checkheap_errors += 1;
+        printf("!!! dangling reference at %s"PHtx" to "PHtx"%s\n",
+               (address_is_meaningful ? "" : "insideof "),
+               (ptrdiff_t)(address_is_meaningful ? pp : TO_VOIDP(base)),
+               (ptrdiff_t)p, (aftergc ? " after gc" : ""));
+        printf("from: "); segment_tell(seg);
+        printf("to:   "); segment_tell(ptr_get_segment(p));
+        {
+          ptr l;
+          for (l = S_G.locked_objects[psi->generation]; l != Snil; l = Scdr(l))
+            if (Scar(l) == p)
+              printf(" in locked\n");
+          for (l = S_G.unlocked_objects[psi->generation]; l != Snil; l = Scdr(l))
+            if (Scar(l) == p)
+              printf(" in unlocked\n");
+        }
+      }
+
+      if (address_is_meaningful) {
+        seginfo *ppsi = MaybeSegInfo(ptr_get_segment(TO_PTR(pp)));
+        if ((ppsi != NULL)
+            && (ppsi->generation > psi->generation)
+            /* space_data includes stacks, which are always swept */
+            && (ppsi->space != space_data)) {
+          uptr card = (uptr)TO_PTR(pp) >> card_offset_bits;
+          uptr cardno = card & ((1 << segment_card_offset_bits) - 1);
+          if (psi->generation < ppsi->dirty_bytes[cardno]) {
+            S_checkheap_errors += 1;
+            check_heap_dirty_msg("!!! INVALID", pp);
+          }
+        }
+      }
+    }
+  }
+}
+
+static void check_bignum(ptr p) {
+  if (!Sbignump(p))
+    printf("!!! not a bignum %p\n", TO_VOIDP(p));
+}
+
+#include "heapcheck.inc"
+
+#ifdef PTHREADS
+
+static ptr *find_nl(ptr *pp1, ISPC s, IGEN g) {
+  seginfo *si;
+  thread_gc *tgc;
+
+  si = SegInfo(addr_get_segment(TO_PTR(pp1)));
+  tgc = si->creator;
+  return TO_VOIDP(tgc->next_loc[g][s]);
+}
+
+
+# define FIND_NL(pp1, pp2, s, g) find_nl(pp1, s, g)
+
+#else
+
+/* no `creator` field, so we search the slow way */
+static ptr *find_nl(ptr *pp1, ptr *pp2, ISPC s, IGEN g) {
+  ptr *nl, ls;
+
+  for (ls = S_threads; ls != Snil; ls = Scdr(ls)) {
+    ptr t_tc = (ptr)THREADTC(Scar(ls));
+    nl = TO_VOIDP(THREAD_GC(t_tc)->next_loc[g][s]);
+    if (pp1 <= nl && nl < pp2)
+      return nl;
+  }
+
+  return NULL;
+}
+# define FIND_NL(pp1, pp2, s, g) find_nl(pp1, pp2, s, g)
+
+#endif
+
 static void check_heap_dirty_msg(msg, x) char *msg; ptr *x; {
     INT d; seginfo *si;
 
-    si = SegInfo(addr_get_segment(x));
-    d = (INT)(((uptr)x >> card_offset_bits) & ((1 << segment_card_offset_bits) - 1));
-    printf("%s dirty byte %d found in segment %#tx, card %d at %#tx\n", msg, si->dirty_bytes[d], (ptrdiff_t)(si->number), d, (ptrdiff_t)x);
-    printf("from "); segment_tell(addr_get_segment(x));
+    si = SegInfo(addr_get_segment(TO_PTR(x)));
+    d = (INT)(((uptr)TO_PTR(x) >> card_offset_bits) & ((1 << segment_card_offset_bits) - 1));
+    printf("%s dirty byte %d found in segment "PHtx", card %d at "PHtx"\n", msg, si->dirty_bytes[d], (ptrdiff_t)(si->number), d, (ptrdiff_t)x);
+    printf("from "); segment_tell(addr_get_segment(TO_PTR(x)));
     printf("to   "); segment_tell(addr_get_segment(*x));
 }
 
@@ -548,6 +654,30 @@ void S_check_heap(aftergc, mcg) IBOOL aftergc; IGEN mcg; {
 
   check_dirty();
 
+  {
+    ptr ls;
+    for (ls = S_threads; ls != Snil; ls = Scdr(ls)) {
+      ptr t_tc = (ptr)THREADTC(Scar(ls));
+      for (s = 0; s <= max_real_space; s += 1) {
+        for (g = 0; g <= static_generation; INCRGEN(g)) {
+          thread_gc *tgc = THREAD_GC(t_tc);
+          if ((tgc->next_loc[g][s] == (ptr)0) != (tgc->base_loc[g][s] == (ptr)0)) {
+            S_checkheap_errors += 1;
+            printf("!!! inconsistent thread NEXT %p and BASE %p\n",
+                   TO_VOIDP(tgc->next_loc[g][s]), TO_VOIDP(tgc->base_loc[g][s]));
+          }
+          if ((tgc->remote_range_end != (ptr)0)
+              || (tgc->remote_range_start != (ptr)(uptr)-1)) {
+            S_checkheap_errors += 1;
+            printf("!!! nonempty thread REMOTERANGE %p-%p\n",
+                   TO_VOIDP(tgc->remote_range_start),
+                   TO_VOIDP(tgc->remote_range_end));
+          }
+        }
+      }
+    }
+  }
+
   for (i = PARTIAL_CHUNK_POOLS; i >= -1; i -= 1) {
     chunkinfo *chunk = i == -1 ? S_chunks_full : S_chunks[i];
     while (chunk != NULL) {
@@ -563,7 +693,7 @@ void S_check_heap(aftergc, mcg) IBOOL aftergc; IGEN mcg; {
       }
       if ((chunk->segs - count) != chunk->nused_segs) {
         S_checkheap_errors += 1;
-        printf("!!! unexpected used segs count %td with %td total segs and %td segs on the unused list\n",
+        printf("!!! unexpected used segs count "Ptd" with "Ptd" total segs and "Ptd" segs on the unused list\n",
             (ptrdiff_t)chunk->nused_segs, (ptrdiff_t)chunk->segs, (ptrdiff_t)count);
       }
       used_segments += chunk->nused_segs;
@@ -575,7 +705,7 @@ void S_check_heap(aftergc, mcg) IBOOL aftergc; IGEN mcg; {
   for (s = 0; s <= max_real_space; s += 1) {
     seginfo *si;
     for (g = 0; g <= S_G.max_nonstatic_generation; INCRGEN(g)) {
-      for (si = S_G.occupied_segments[s][g]; si != NULL; si = si->next) {
+      for (si = S_G.occupied_segments[g][s]; si != NULL; si = si->next) {
         if (si->generation != g) {
           S_checkheap_errors += 1;
           printf("!!! segment in wrong occupied_segments list\n");
@@ -583,28 +713,28 @@ void S_check_heap(aftergc, mcg) IBOOL aftergc; IGEN mcg; {
         nonstatic_segments += 1;
       }
     }
-    for (si = S_G.occupied_segments[s][static_generation]; si != NULL; si = si->next) {
+    for (si = S_G.occupied_segments[static_generation][s]; si != NULL; si = si->next) {
       static_segments += 1;
     }
   }
 
   if (used_segments != nonstatic_segments + static_segments) {
     S_checkheap_errors += 1;
-    printf("!!! found %#tx used segments and %#tx occupied segments\n",
+    printf("!!! found "PHtx" used segments and "PHtx" occupied segments\n",
         (ptrdiff_t)used_segments,
         (ptrdiff_t)(nonstatic_segments + static_segments));
   }
 
   if (S_G.number_of_nonstatic_segments != nonstatic_segments) {
     S_checkheap_errors += 1;
-    printf("!!! S_G.number_of_nonstatic_segments %#tx is different from occupied number %#tx\n",
+    printf("!!! S_G.number_of_nonstatic_segments "PHtx" is different from occupied number "PHtx"\n",
         (ptrdiff_t)S_G.number_of_nonstatic_segments,
         (ptrdiff_t)nonstatic_segments);
   }
 
   if (S_G.number_of_empty_segments != empty_segments) {
     S_checkheap_errors += 1;
-    printf("!!! S_G.number_of_empty_segments %#tx is different from unused number %#tx\n",
+    printf("!!! S_G.number_of_empty_segments "PHtx" is different from unused number "PHtx"\n",
         (ptrdiff_t)S_G.number_of_empty_segments,
         (ptrdiff_t)empty_segments);
   }
@@ -617,87 +747,172 @@ void S_check_heap(aftergc, mcg) IBOOL aftergc; IGEN mcg; {
         seginfo *recorded_si; uptr recorded_seg;
         if ((seg = si->number) != (recorded_seg = (chunk->base + chunk->segs - nsegs))) {
           S_checkheap_errors += 1;
-          printf("!!! recorded segment number %#tx differs from actual segment number %#tx", (ptrdiff_t)seg, (ptrdiff_t)recorded_seg);
+          printf("!!! recorded segment number "PHtx" differs from actual segment number "PHtx"", (ptrdiff_t)seg, (ptrdiff_t)recorded_seg);
         }
         if ((recorded_si = SegInfo(seg)) != si) {
           S_checkheap_errors += 1;
-          printf("!!! recorded segment %#tx seginfo %#tx differs from actual seginfo %#tx", (ptrdiff_t)seg, (ptrdiff_t)recorded_si, (ptrdiff_t)si);
+          printf("!!! recorded segment "PHtx" seginfo "PHtx" differs from actual seginfo "PHtx"", (ptrdiff_t)seg, (ptrdiff_t)recorded_si, (ptrdiff_t)si);
         }
         s = si->space;
         g = si->generation;
 
         if (si->use_marks)
-          printf("!!! use_marks set on generation %d segment %#tx\n", g, (ptrdiff_t)seg);
+          printf("!!! use_marks set on generation %d segment "PHtx"\n", g, (ptrdiff_t)seg);
 
         if (s == space_new) {
           if (g != 0 && !si->marked_mask) {
             S_checkheap_errors += 1;
-            printf("!!! unexpected generation %d segment %#tx in space_new\n", g, (ptrdiff_t)seg);
+            printf("!!! unexpected generation %d segment "PHtx" in space_new\n", g, (ptrdiff_t)seg);
           }
         } else if (s == space_impure || s == space_symbol || s == space_pure || s == space_weakpair || s == space_ephemeron
-                   || s == space_immobile_impure || s == space_count_pure || s == space_count_impure || s == space_closure) {
-          /* doesn't handle: space_port, space_continuation, space_code, space_pure_typed_object,
-                             space_impure_record, or impure_typed_object */
-          nl = (ptr *)S_G.next_loc[s][g];
-
+                   || s == space_immobile_impure || s == space_count_pure || s == space_count_impure || s == space_closure
+                   || s == space_pure_typed_object || s == space_continuation || s == space_port || s == space_code
+                   || s == space_impure_record || s == space_impure_typed_object) {
+          ptr start;
+          
           /* check for dangling references */
-          pp1 = (ptr *)build_ptr(seg, 0);
-          pp2 = (ptr *)build_ptr(seg + 1, 0);
+          pp1 = TO_VOIDP(build_ptr(seg, 0));
+          pp2 = TO_VOIDP(build_ptr(seg + 1, 0));
+
+          nl = FIND_NL(pp1, pp2, s, g);
           if (pp1 <= nl && nl < pp2) pp2 = nl;
 
-          while (pp1 < pp2) {
-            if (!si->marked_mask || (si->marked_mask[segment_bitmap_byte(pp1)] & segment_bitmap_bit(pp1))) {
-              int a;
-              for (a = 0; (a < ptr_alignment) && (pp1 < pp2); a++) {
-#define         in_ephemeron_pair_part(pp1, seg) ((((uptr)(pp1) - (uptr)build_ptr(seg, 0)) % size_ephemeron) < size_pair)
-                if ((s == space_ephemeron) && !in_ephemeron_pair_part(pp1, seg)) {
-                  /* skip non-pair part of ephemeron */
+          if (s == space_pure_typed_object || s == space_port || s == space_code
+              || s == space_impure_record || s == space_impure_typed_object) {
+            /* only check this segment for objects that start on it */
+            uptr before_seg = seg;
+
+            /* Back up over segments for the same space and generation: */
+            while (1) {
+              seginfo *before_si = MaybeSegInfo(before_seg-1);
+              if (!before_si
+                  || (before_si->space != si->space)
+                  || (before_si->generation != si->generation)
+                  || ((before_si->marked_mask == NULL) != (si->marked_mask == NULL)))
+                break;
+              before_seg--;
+            }
+
+            /* Move forward to reach `seg` again: */
+            start = build_ptr(before_seg, 0);
+            while (before_seg != seg) {
+              ptr *before_pp2, *before_nl;
+
+              before_pp2 = TO_VOIDP(build_ptr(before_seg + 1, 0));
+              if ((ptr *)TO_VOIDP(start) > before_pp2) {
+                /* skipped to a further segment */
+                before_seg++;
+              } else {
+                before_nl = FIND_NL(TO_VOIDP(start), before_pp2, s, g);
+                if (((ptr*)TO_VOIDP(start)) <= before_nl && before_nl < before_pp2) {
+                  /* this segment ends, so move to next segment */
+                  before_seg++;
+                  if (s == space_code) {
+                    /* in the case of code, it's possible for a whole segment to
+                       go unused if a large code object didn't fit; give up, just in case */
+                    start = build_ptr(seg+1, 0);
+                  } else {
+                    start = build_ptr(before_seg, 0);
+                  }
                 } else {
-                  p = *pp1;
-                  if (p == forward_marker) {
-                    pp1 = pp2; /* break out of outer loop */
-                    break;
-                  } else if (!IMMEDIATE(p)) {
-                    seginfo *psi = MaybeSegInfo(ptr_get_segment(p));
-                    if (psi != NULL) {
-                      if ((psi->space == space_empty)
-                          || psi->old_space
-                          || (psi->marked_mask && !(psi->marked_mask[segment_bitmap_byte(p)] & segment_bitmap_bit(p)))) {
-                        S_checkheap_errors += 1;
-                        printf("!!! dangling reference at %#tx to %#tx%s\n", (ptrdiff_t)pp1, (ptrdiff_t)p, (aftergc ? " after gc" : ""));
-                        printf("from: "); segment_tell(seg);
-                        printf("to:   "); segment_tell(ptr_get_segment(p));
-                        {
-                          ptr l;
-                          for (l = S_G.locked_objects[psi->generation]; l != Snil; l = Scdr(l))
-                            if (Scar(l) == p)
-                              printf(" in locked\n");
-                          for (l = S_G.unlocked_objects[psi->generation]; l != Snil; l = Scdr(l))
-                            if (Scar(l) == p)
-                              printf(" in unlocked\n");
+                  seginfo *before_si = MaybeSegInfo(before_seg);
+                  while (((ptr *)TO_VOIDP(start)) < before_pp2) {
+                    if (before_si->marked_mask) {
+                      if (before_si->marked_mask[segment_bitmap_byte(start)] & segment_bitmap_bit(start)) {
+                        start = (ptr)((uptr)start + size_object(TYPE(start, type_typed_object)));
+                      } else {
+                        /* skip past unmarked */
+                        start = (ptr)((uptr)start + byte_alignment);
+                      }
+                    } else {
+                      if (*(ptr *)TO_VOIDP(start) == forward_marker) {
+                        /* this segment ends, so move to next segment */
+                        if (s == space_code) {
+                          start = build_ptr(seg+1, 0);
+                        } else {
+                          start = build_ptr(before_seg+1, 0);
                         }
+                      } else {
+                        start = (ptr)((uptr)start + size_object(TYPE(start, type_typed_object)));
                       }
                     }
                   }
+                  before_seg++;
                 }
-                pp1 += 1;
               }
-            } else
-              pp1 += ptr_alignment;
+            }
+
+            if (((ptr *)TO_VOIDP(start)) >= pp2) {
+              /* previous object extended past the segment */
+            }  else {
+              pp1 = TO_VOIDP(start);
+              while (pp1 < pp2) {
+                if (si->marked_mask) {
+                  if (si->marked_mask[segment_bitmap_byte(TO_PTR(pp1))] & segment_bitmap_bit(TO_PTR(pp1))) {
+                    p = TYPE(TO_PTR(pp1), type_typed_object);
+                    check_object(p, seg, s, aftergc);
+                    pp1 = TO_VOIDP((ptr)((uptr)TO_PTR(pp1) + size_object(p)));
+                  } else {
+                    /* skip past unmarked */
+                    pp1 = TO_VOIDP((ptr)((uptr)TO_PTR(pp1) + byte_alignment));
+                  }
+                } else {
+                  if (*pp1 == forward_marker)
+                    break;
+                  else {
+                    p = TYPE(TO_PTR(pp1), type_typed_object);
+                    check_object(p, seg, s, aftergc);
+                    pp1 = TO_VOIDP((ptr)((uptr)TO_PTR(pp1) + size_object(p)));
+                  }
+                }
+              }
+            }
+          } else if (s == space_continuation) {
+            while (pp1 < pp2) {
+              if (*pp1 == forward_marker)
+                break;
+              if (!si->marked_mask || (si->marked_mask[segment_bitmap_byte(TO_PTR(pp1))] & segment_bitmap_bit(TO_PTR(pp1)))) {
+                p = TYPE(TO_PTR(pp1), type_closure);
+                check_object(p, seg, s, aftergc);
+              }
+              pp1 = TO_VOIDP((ptr)((uptr)TO_PTR(pp1) + size_continuation));
+            }
+          } else {
+            while (pp1 < pp2) {
+              if (!si->marked_mask || (si->marked_mask[segment_bitmap_byte(TO_PTR(pp1))] & segment_bitmap_bit(TO_PTR(pp1)))) {
+                int a;
+                for (a = 0; (a < ptr_alignment) && (pp1 < pp2); a++) {
+#define         in_ephemeron_pair_part(pp1, seg) ((((uptr)TO_PTR(pp1) - (uptr)build_ptr(seg, 0)) % size_ephemeron) < size_pair)
+                  if ((s == space_ephemeron) && !in_ephemeron_pair_part(pp1, seg)) {
+                    /* skip non-pair part of ephemeron */
+                  } else {
+                    p = *pp1;
+                    if (!si->marked_mask && (p == forward_marker)) {
+                      pp1 = pp2; /* break out of outer loop */
+                      break;
+                    } else {
+                      check_pointer(pp1, 1, (ptr)0, seg, s, aftergc);
+                    }
+                  }
+                  pp1 += 1;
+                }
+              } else
+                pp1 += ptr_alignment;
+            }
           }
 
-          /* verify that dirty bits are set appropriately */
-          /* out of date: doesn't handle space_impure_record, space_port, and maybe others */
-          /* also doesn't check the SYMCODE for symbols */
+          /* further verify that dirty bits are set appropriately; only handles some spaces
+             to make sure that the dirty byte is not unnecessarily approximate, but we have also
+             checked dirty bytes alerady via `check_pointer` */
           if (s == space_impure || s == space_symbol || s == space_weakpair || s == space_ephemeron
               || s == space_immobile_impure || s == space_closure) {
             found_eos = 0;
-            pp2 = pp1 = build_ptr(seg, 0);
+            pp2 = pp1 = TO_VOIDP(build_ptr(seg, 0));
             for (d = 0; d < cards_per_segment; d += 1) {
               if (found_eos) {
                 if (si->dirty_bytes[d] != 0xff) {
                   S_checkheap_errors += 1;
-                  printf("!!! Dirty byte set past end-of-segment for segment %#tx, card %d\n", (ptrdiff_t)seg, d);
+                  printf("!!! Dirty byte set past end-of-segment for segment "PHtx", card %d\n", (ptrdiff_t)seg, d);
                   segment_tell(seg);
                 }
                 continue;
@@ -710,13 +925,13 @@ void S_check_heap(aftergc, mcg) IBOOL aftergc; IGEN mcg; {
               }
 
 #ifdef DEBUG
-              printf("pp1 = %#tx, pp2 = %#tx, nl = %#tx\n", (ptrdiff_t)pp1, (ptrdiff_t)pp2, (ptrdiff_t)nl);
+              printf("pp1 = "PHtx", pp2 = "PHtx", nl = "PHtx"\n", (ptrdiff_t)pp1, (ptrdiff_t)pp2, (ptrdiff_t)nl);
               fflush(stdout);
 #endif
 
               dirty = 0xff;
               while (pp1 < pp2) {
-                if (!si->marked_mask || (si->marked_mask[segment_bitmap_byte(pp1)] & segment_bitmap_bit(pp1))) {
+                if (!si->marked_mask || (si->marked_mask[segment_bitmap_byte(TO_PTR(pp1))] & segment_bitmap_bit(TO_PTR(pp1)))) {
                   int a;
                   for (a = 0; (a < ptr_alignment) && (pp1 < pp2); a++) {
                     if ((s == space_ephemeron) && !in_ephemeron_pair_part(pp1, seg)) {
@@ -752,13 +967,27 @@ void S_check_heap(aftergc, mcg) IBOOL aftergc; IGEN mcg; {
                    cards with dirty pointers to segments older than the
                    maximum copyied generation, so we can get legitimate
                    conservative dirty bytes even after gc */
-                printf("... Conservative dirty byte %x (%x) %sfor segment %#tx card %d ",
+                printf("... Conservative dirty byte %x (%x) %sfor segment "PHtx" card %d ",
                        si->dirty_bytes[d], dirty,
                        (aftergc ? "after gc " : ""),
                        (ptrdiff_t)seg, d);
                 segment_tell(seg);
               }
             }
+          } else {
+            /* at least check that no dirty bytes are set beyond the end of the segment */
+            if (pp2 < (ptr *)TO_VOIDP(build_ptr(seg + 1, 0))) {
+              uptr card = (uptr)TO_PTR(pp2) >> card_offset_bits;
+              int d = (int)(card & ((1 << segment_card_offset_bits) - 1));
+
+              for (d++; d < cards_per_segment; d++) {
+                if (si->dirty_bytes[d] != 0xff) {
+                  S_checkheap_errors += 1;
+                  printf("!!! Dirty byte set past end-of-segment for segment "PHtx", card %d\n", (ptrdiff_t)seg, d);
+                  segment_tell(seg);
+                }
+              }
+            }
           }
         }
         if (aftergc
@@ -770,7 +999,7 @@ void S_check_heap(aftergc, mcg) IBOOL aftergc; IGEN mcg; {
           for (d = 0; d < cards_per_segment; d += 1) {
             if (si->dirty_bytes[d] != 0xff) {
               S_checkheap_errors += 1;
-              printf("!!! Unnecessary dirty byte %x (%x) after gc for segment %#tx card %d ",
+              printf("!!! Unnecessary dirty byte %x (%x) after gc for segment "PHtx" card %d ",
                   si->dirty_bytes[d], 0xff, (ptrdiff_t)(si->number), d);
               segment_tell(seg);
             }
@@ -810,7 +1039,7 @@ static void check_dirty_space(ISPC s) {
   IGEN from_g, to_g, min_to_g; INT d; seginfo *si;
 
   for (from_g = 0; from_g <= static_generation; from_g += 1) {
-    for (si = S_G.occupied_segments[s][from_g]; si != NULL; si = si->next) {
+    for (si = S_G.occupied_segments[from_g][s]; si != NULL; si = si->next) {
       min_to_g = 0xff;
       for (d = 0; d < cards_per_segment; d += 1) {
         to_g = si->dirty_bytes[d];
@@ -818,18 +1047,18 @@ static void check_dirty_space(ISPC s) {
           if (to_g < min_to_g) min_to_g = to_g;
           if (from_g == 0) {
             S_checkheap_errors += 1;
-            printf("!!! (check_dirty): space %d, generation %d segment %#tx card %d is marked dirty\n", s, from_g, (ptrdiff_t)(si->number), d);
+            printf("!!! (check_dirty): space %d, generation %d segment "PHtx" card %d is marked dirty\n", s, from_g, (ptrdiff_t)(si->number), d);
           }
         }
       }
       if (min_to_g != si->min_dirty_byte) {
         S_checkheap_errors += 1;
-        printf("!!! (check_dirty): space %d, generation %d segment %#tx min_dirty_byte is %d while actual min is %d\n",  s, from_g, (ptrdiff_t)(si->number), si->min_dirty_byte, min_to_g);
+        printf("!!! (check_dirty): space %d, generation %d segment "PHtx" min_dirty_byte is %d while actual min is %d\n",  s, from_g, (ptrdiff_t)(si->number), si->min_dirty_byte, min_to_g);
         segment_tell(si->number);
       } else if (min_to_g != 0xff) {
         if (!dirty_listedp(si, from_g, min_to_g)) {
           S_checkheap_errors += 1;
-          printf("!!! (check_dirty): space %d, generation %d segment %#tx is marked dirty but not in dirty-segment list\n", s, from_g, (ptrdiff_t)(si->number));
+          printf("!!! (check_dirty): space %d, generation %d segment "PHtx" is marked dirty but not in dirty-segment list\n", s, from_g, (ptrdiff_t)(si->number));
           segment_tell(si->number);
         }
       }
@@ -855,17 +1084,17 @@ static void check_dirty() {
           IGEN mingval = si->min_dirty_byte;
           if (g != from_g) {
             S_checkheap_errors += 1;
-            printf("!!! (check_dirty): generation %d segment %#tx in %d -> %d dirty list\n", g, (ptrdiff_t)(si->number), from_g, to_g);
+            printf("!!! (check_dirty): generation %d segment "PHtx" in %d -> %d dirty list\n", g, (ptrdiff_t)(si->number), from_g, to_g);
           }
           if (mingval != to_g) {
             S_checkheap_errors += 1;
-            printf("!!! (check_dirty): dirty byte = %d for segment %#tx in %d -> %d dirty list\n", mingval, (ptrdiff_t)(si->number), from_g, to_g);
+            printf("!!! (check_dirty): dirty byte = %d for segment "PHtx" in %d -> %d dirty list\n", mingval, (ptrdiff_t)(si->number), from_g, to_g);
           }
-          if (s != space_new && s != space_impure && s != space_symbol && s != space_port
+          if (s != space_new && s != space_impure && s != space_count_impure && s != space_symbol && s != space_port
               && s != space_impure_record && s != space_impure_typed_object && s != space_immobile_impure 
-              && s != space_weakpair && s != space_ephemeron) {
+              && s != space_weakpair && s != space_ephemeron && s != space_closure) {
             S_checkheap_errors += 1;
-            printf("!!! (check_dirty): unexpected space %d for dirty segment %#tx\n", s, (ptrdiff_t)(si->number));
+            printf("!!! (check_dirty): unexpected space %d for dirty segment "PHtx"\n", s, (ptrdiff_t)(si->number));
           }
           si = si->dirty_next;
         }
@@ -890,25 +1119,25 @@ static void check_locked_object(ptr p, IBOOL locked, IGEN g, IBOOL aftergc, IGEN
   seginfo *psi = MaybeSegInfo(ptr_get_segment(p));
   if (!psi) {
     S_checkheap_errors += 1;
-    printf("!!! generation %d %s object has no segment: %p\n", g, what, p);
+    printf("!!! generation %d %s object has no segment: %p\n", g, what, TO_VOIDP(p));
   } else {
     if (psi->generation != g) {
       S_checkheap_errors += 1;
-      printf("!!! generation %d %s object in generation %d segment: %p\n", g, what, psi->generation, p);
+      printf("!!! generation %d %s object in generation %d segment: %p\n", g, what, psi->generation, TO_VOIDP(p));
     }
     if (!psi->must_mark && locked) {
       S_checkheap_errors += 1;
-      printf("!!! generation %d %s object not on must-mark page: %p\n", g, what, p);
+      printf("!!! generation %d %s object not on must-mark page: %p\n", g, what, TO_VOIDP(p));
     }
     if (!psi->marked_mask) {
       if (aftergc && (psi->generation <= mcg)) {
         S_checkheap_errors += 1;
-        printf("!!! %s object not in marked segment: %p\n", what, p);
+        printf("!!! %s object not in marked segment: %p\n", what, TO_VOIDP(p));
         printf(" in: "); segment_tell(psi->number);
       }
     } else if (!(psi->marked_mask[segment_bitmap_byte(p)] & segment_bitmap_bit(p))) {
       S_checkheap_errors += 1;
-      printf("!!! generation %d %s object not marked: %p\n", g, what, p);
+      printf("!!! generation %d %s object not marked: %p\n", g, what, TO_VOIDP(p));
     }
   }
 }
@@ -924,7 +1153,7 @@ void S_fixup_counts(ptr counts) {
   RTDCOUNTSTIMESTAMP(counts) = S_G.gctimestamp[0];
 }
 
-ptr S_do_gc(IGEN mcg, IGEN tg, ptr count_roots) {
+ptr S_do_gc(IGEN max_cg, IGEN min_tg, IGEN max_tg, ptr count_roots) {
   ptr tc = get_thread_context();
   ptr code, result;
 
@@ -940,24 +1169,33 @@ ptr S_do_gc(IGEN mcg, IGEN tg, ptr count_roots) {
     S_G.max_nonstatic_generation = S_G.new_max_nonstatic_generation;
   }
 
-  if (tg == mcg && mcg == S_G.new_max_nonstatic_generation && mcg < S_G.max_nonstatic_generation) {
+  if (max_tg == max_cg && max_cg == S_G.new_max_nonstatic_generation && max_cg < S_G.max_nonstatic_generation) {
     IGEN new_g, old_g, from_g, to_g; ISPC s; seginfo *si, *nextsi, *tail;
    /* reducing max_nonstatic_generation */
     new_g = S_G.new_max_nonstatic_generation;
     old_g = S_G.max_nonstatic_generation;
-   /* first, collect everything to old_g */
-    result = S_gc(tc, old_g, old_g, count_roots);
+   /* first, collect everything to old_g, ignoring min_tg */
+    result = S_gc(tc, old_g, old_g, old_g, count_roots);
    /* now transfer old_g info to new_g, and clear old_g info */
+    S_G.bytes_of_generation[new_g] = S_G.bytes_of_generation[old_g]; S_G.bytes_of_generation[old_g] = 0;
     for (s = 0; s <= max_real_space; s += 1) {
-      S_G.first_loc[s][new_g] = S_G.first_loc[s][old_g]; S_G.first_loc[s][old_g] = FIX(0);
-      S_G.base_loc[s][new_g] = S_G.base_loc[s][old_g]; S_G.base_loc[s][old_g] = FIX(0);
-      S_G.next_loc[s][new_g] = S_G.next_loc[s][old_g]; S_G.next_loc[s][old_g] = FIX(0);
-      S_G.bytes_left[s][new_g] = S_G.bytes_left[s][old_g]; S_G.bytes_left[s][old_g] = 0;
-      S_G.bytes_of_space[s][new_g] = S_G.bytes_of_space[s][old_g]; S_G.bytes_of_space[s][old_g] = 0;
-      S_G.occupied_segments[s][new_g] = S_G.occupied_segments[s][old_g]; S_G.occupied_segments[s][old_g] = NULL;
-      for (si = S_G.occupied_segments[s][new_g]; si != NULL; si = si->next) {
+      S_G.bytes_of_space[new_g][s] = S_G.bytes_of_space[old_g][s]; S_G.bytes_of_space[old_g][s] = 0;
+      S_G.occupied_segments[new_g][s] = S_G.occupied_segments[old_g][s]; S_G.occupied_segments[old_g][s] = NULL;
+      for (si = S_G.occupied_segments[new_g][s]; si != NULL; si = si->next) {
         si->generation = new_g;
       }
+      {
+        ptr ls;
+        for (ls = S_threads; ls != Snil; ls = Scdr(ls)) {
+          ptr t_tc = (ptr)THREADTC(Scar(ls));
+          thread_gc *tgc = THREAD_GC(t_tc);
+          tgc->base_loc[new_g][s] = tgc->base_loc[old_g][s]; tgc->base_loc[old_g][s] = (ptr)0;
+          tgc->next_loc[new_g][s] = tgc->next_loc[old_g][s]; tgc->next_loc[old_g][s] = (ptr)0;
+          tgc->bytes_left[new_g][s] = tgc->bytes_left[old_g][s]; tgc->bytes_left[old_g][s] = 0;
+          tgc->sweep_loc[new_g][s] = tgc->sweep_loc[old_g][s]; tgc->sweep_loc[old_g][s] = 0;
+          tgc->sweep_next[new_g][s] = tgc->sweep_next[old_g][s]; tgc->sweep_next[old_g][s] = NULL;
+        }
+      }
     }
     S_G.guardians[new_g] = S_G.guardians[old_g]; S_G.guardians[old_g] = Snil;
     S_G.locked_objects[new_g] = S_G.locked_objects[old_g]; S_G.locked_objects[old_g] = Snil;
@@ -979,6 +1217,9 @@ ptr S_do_gc(IGEN mcg, IGEN tg, ptr count_roots) {
         RTDCOUNTSIT(counts, new_g) = RTDCOUNTSIT(counts, old_g); RTDCOUNTSIT(counts, old_g) = 0;
       }
     }
+#ifndef WIN32
+    S_child_processes[new_g] = S_child_processes[old_g];
+#endif
 
     /* change old_g dirty bytes in static generation to new_g; splice list of old_g
        seginfos onto front of new_g seginfos */
@@ -1030,25 +1271,37 @@ ptr S_do_gc(IGEN mcg, IGEN tg, ptr count_roots) {
     S_G.min_free_gen = S_G.new_min_free_gen;
     S_G.max_nonstatic_generation = new_g;
   } else {
-    result = S_gc(tc, mcg, tg, count_roots);
+    result = S_gc(tc, max_cg, min_tg, max_tg, count_roots);
   }
-  S_pants_down -= 1;
 
  /* eagerly give collecting thread, the only one guaranteed to be
     active, a fresh allocation area.  the other threads have to trap
     to get_more_room if and when they awake and try to allocate */
   S_reset_allocation_pointer(tc);
 
+  S_pants_down -= 1;
+
   Sunlock_object(code);
 
   return result;
 }
 
-ptr S_gc(ptr tc, IGEN mcg, IGEN tg, ptr count_roots) {
-  if (tg == static_generation
+ptr S_gc(ptr tc, IGEN max_cg, IGEN min_tg, IGEN max_tg, ptr count_roots) {
+  if (min_tg == static_generation
       || S_G.enable_object_counts || S_G.enable_object_backreferences
-      || (count_roots != Sfalse))
-    return S_gc_oce(tc, mcg, tg, count_roots);
-  else
-    return S_gc_ocd(tc, mcg, tg, Sfalse);
+      || (count_roots != Sfalse)) {
+    if (S_G.enable_object_backreferences) min_tg = max_tg;
+    return S_gc_oce(tc, max_cg, min_tg, max_tg, count_roots);
+#if defined(PTHREADS)
+  } else if (S_collect_waiting_threads != 0) {
+    return S_gc_par(tc, max_cg, min_tg, max_tg, Sfalse);
+#endif
+  } else if (max_cg == 0 && min_tg == 1 && max_tg == 1
+           && !S_G.must_mark_gen0 && S_G.locked_objects[0] == Snil
+           && (S_G.min_mark_gen > 0)) {
+    S_gc_011(tc);
+    return Svoid;
+  } else {
+    return S_gc_ocd(tc, max_cg, min_tg, max_tg, Sfalse);
+  }
 }
