@@ -38,6 +38,8 @@ rktio_t *rktio_init(void)
 
   rktio_syslog_init(rktio);
 
+  rktio_convert_init(rktio);
+
 #ifdef OS_X
   {
     int a[2], i, k = 0;
@@ -66,12 +68,14 @@ void rktio_destroy(rktio_t *rktio)
 {
   rktio_stop_background(rktio);
   rktio_syslog_clean(rktio);
+  rktio_convert_deinit(rktio);
   rktio_dll_clean(rktio);
   rktio_error_clean(rktio);
   rktio_process_deinit(rktio);
   rktio_free_ghbn(rktio);
   rktio_free_global_poll_set(rktio);
   rktio_stop_fs_change(rktio);
+  rktio_free_signal(rktio);
 #ifdef RKTIO_SYSTEM_WINDOWS
   rktio_winsock_done(rktio);
 #endif
